# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import logging

from fontTools.varLib.models import piecewiseLinearMap

from glyphsLib.util import build_ufo_path
from glyphsLib.classes import (
    CustomParametersProxy,
    GSCustomParameter,
    InstanceType,
    PropertiesProxy,
    WEIGHT_CODES,
)
from .constants import (
    UFO_FILENAME_CUSTOM_PARAM,
    EXPORT_KEY,
    WIDTH_KEY,
    WEIGHT_KEY,
    FULL_FILENAME_KEY,
    MANUAL_INTERPOLATION_KEY,
    INSTANCE_INTERPOLATIONS_KEY,
    CUSTOM_PARAMETERS_KEY,
    CUSTOM_PARAMETERS_BLACKLIST,
    PROPERTIES_KEY,
    PROPERTIES_WHITELIST,
)
from .names import build_stylemap_names
from .axes import (
    get_axis_definitions,
    is_instance_active,
    WEIGHT_AXIS_DEF,
    WIDTH_AXIS_DEF,
    AxisDefinitionFactory,
)
from .custom_params import to_ufo_custom_params


logger = logging.getLogger(__name__)


def to_designspace_instances(self):
    """Write instance data from self.font to self.designspace."""
    for instance in self.font.instances:
        if self.minimize_glyphs_diffs or (
            is_instance_active(instance)
            and _is_instance_included_in_family(self, instance)
        ):
            if instance.type == InstanceType.VARIABLE:
                _to_designspace_varfont(self, instance)
            else:
                # In 'minimal' mode (enabled e.g. by fontmake when converting
                # .glyphs => .ufo for compiling OpenType fonts) we treat disabled
                # custom parameters as if they are absent, thus here we skip
                # exporting them to the designspace instance when minimal=True.
                _to_designspace_instance(
                    self, instance, ignore_disabled_cp=self.minimal
                )


def _to_designspace_varfont(self, instance):
    from fontTools.designspaceLib import RangeAxisSubsetDescriptor
    from fontTools.ufoLib import fontInfoAttributesVersion3

    ds = self.designspace

    # unless the `fileName` custom parameter was explicitly set, do like Glyphs.app
    # and concatenate the family, instance (style) names and "VF" to form a filename;
    # the default 'Regular' is omitted by Glyphs.app.
    # https://github.com/googlefonts/glyphsLib/issues/981
    filename = instance.customParameters["fileName"]
    if not filename:
        filename = self.font.familyName
        if instance.name != "Regular":
            filename += "-" + instance.name
        filename += "VF"
        filename = filename.replace(" ", "")

    ufo_varfont = self.designspace.addVariableFontDescriptor(
        name=instance.name,
        filename=filename,
        axisSubsets=[RangeAxisSubsetDescriptor(name=axis.name) for axis in ds.axes],
    )

    # to_ufo_custom_params() wants a UFO, create a dummy one
    ufo = self.ufo_module.Font()
    to_ufo_custom_params(self, ufo, instance, set_default_params=False)

    info = {}
    for attr in fontInfoAttributesVersion3:
        if (value := getattr(ufo.info, attr, None)) is not None:
            info[attr] = value

    if info:
        ufo_varfont.lib["public.fontInfo"] = info

    for key in ufo.lib:
        ufo_varfont.lib[key] = ufo.lib[key]


def _to_designspace_instance(self, instance, ignore_disabled_cp=False):
    ufo_instance = self.designspace.newInstanceDescriptor()

    # FIXME: (jany) most of these customParameters are actually attributes,
    # at least according to https://docu.glyphsapp.com/#fontName

    # Read either from properties or custom parameters or the font
    ufo_instance.familyName = instance.familyName
    ufo_instance.styleName = instance.name
    ufo_instance.postScriptFontName = (
        instance.properties.get("variablePostscriptFontName")
        or instance.properties.get("postscriptFontName")
        or instance.customParameters["postscriptFontName"]
    )
    ufo_instance.filename = _to_filename(self, instance, ufo_instance)

    designspace_axis_tags = {a.tag for a in self.designspace.axes}
    location = {}
    for axis_def in get_axis_definitions(self.font):
        # Only write locations along defined axes
        if axis_def.tag in designspace_axis_tags:
            location[axis_def.name] = axis_def.get_design_loc(instance)
    ufo_instance.location = location

    # FIXME: (jany) should be the responsibility of ufo2ft?
    # Anyway, only generate the styleMap names if the Glyphs instance already
    # has a linkStyle set up, or if we're not round-tripping (i.e. generating
    # UFOs for fontmake, the traditional use-case of glyphsLib.)
    if instance.linkStyle or not self.minimize_glyphs_diffs:
        styleMapFamilyName, styleMapStyleName = build_stylemap_names(
            family_name=ufo_instance.familyName,
            style_name=ufo_instance.styleName,
            is_bold=instance.isBold,
            is_italic=instance.isItalic,
            linked_style=instance.linkStyle,
        )
        ufo_instance.styleMapFamilyName = styleMapFamilyName
        ufo_instance.styleMapStyleName = styleMapStyleName

    ufo_instance.name = " ".join(
        (ufo_instance.familyName or "", ufo_instance.styleName or "")
    )

    if self.minimize_glyphs_diffs:
        ufo_instance.lib[EXPORT_KEY] = instance.active
        ufo_instance.lib[WEIGHT_KEY] = instance.weight
        ufo_instance.lib[WIDTH_KEY] = instance.width

        ufo_instance.lib[INSTANCE_INTERPOLATIONS_KEY] = instance.instanceInterpolations
        ufo_instance.lib[MANUAL_INTERPOLATION_KEY] = instance.manualInterpolation

    # Dump selected custom parameters and properties into the instance
    # descriptor. Later, when using `apply_instance_data`, we will dig out those
    # custom parameters and apply them to the UFO instance.
    parameters = _to_custom_parameters(instance, ignore_disabled_cp)
    if parameters:
        ufo_instance.lib[CUSTOM_PARAMETERS_KEY] = parameters
    properties = _to_properties(instance)
    if properties:
        ufo_instance.lib[PROPERTIES_KEY] = properties

    self.designspace.addInstance(ufo_instance)


def _to_custom_parameters(instance, ignore_disabled=False):
    return [
        (item.name, item.value)
        for item in instance.customParameters
        if item.name not in CUSTOM_PARAMETERS_BLACKLIST
        and not (ignore_disabled and item.disabled)
    ]


def _to_filename(self, instance, ufo_instance):
    filename = (
        instance.customParameters[UFO_FILENAME_CUSTOM_PARAM]
        or instance.customParameters[FULL_FILENAME_KEY]
    )
    if filename:
        if self.instance_dir:
            filename = os.path.basename(filename)
            filename = os.path.join(self.instance_dir, filename)
        return filename
    filename = instance.customParameters["fileName"]
    if filename:
        filename = f"{filename}.ufo"
        if self.instance_dir:
            filename = os.path.join(self.instance_dir, filename)
        return filename
    return build_ufo_path(
        self.instance_dir or "instance_ufos",
        ufo_instance.familyName,
        ufo_instance.styleName,
    )


def _to_properties(instance):
    return [
        (item.name, item.value)
        for item in instance.properties
        if item.name in PROPERTIES_WHITELIST
    ]


def _is_instance_included_in_family(self, instance):
    if not self._do_filter_instances_by_family:
        return True
    return instance.familyName == self.family_name


# TODO: function is too complex (35), split it up
def to_glyphs_instances(self):  # noqa: C901
    if self.designspace is None:
        return

    for ufo_instance in self.designspace.instances:
        instance = self.glyphs_module.GSInstance()

        try:
            instance.active = ufo_instance.lib[EXPORT_KEY]
        except KeyError:
            # If not specified, the default is to export all instances
            instance.active = True

        instance.name = ufo_instance.styleName

        for axis_def in get_axis_definitions(self.font):
            design_loc = None
            try:
                design_loc = ufo_instance.location[axis_def.name]
                axis_def.set_design_loc(instance, design_loc)
            except KeyError:
                # The location does not have this axis?
                pass

            if axis_def.tag in ("wght", "wdth"):
                # Retrieve the user location (weightClass/widthClass)
                # Generic way: read the axis mapping backwards.
                user_loc = design_loc
                mapping = None
                for axis in self.designspace.axes:
                    if axis.tag == axis_def.tag:
                        mapping = axis.map
                if mapping:
                    reverse_mapping = {dl: ul for ul, dl in mapping}
                    user_loc = piecewiseLinearMap(design_loc, reverse_mapping)
                if user_loc is not None:
                    axis_def.set_user_loc(instance, user_loc)

        try:
            # Restore the original weight name when there is an ambiguity based
            # on the value, e.g. Thin, ExtraLight, UltraLight all map to 250.
            # No problem with width, because 1:1 mapping in WIDTH_CODES.
            weight = ufo_instance.lib[WEIGHT_KEY]
            # Only use the lib value if:
            # 1. we don't have a weight for the instance already
            # 2. the value from lib is not "stale", i.e. it still maps to
            #    the current userLocation of the instance. This is in case the
            #    user changes the instance location of the instance by hand but
            #    does not update the weight value in lib.
            if (
                not instance.weight
                or WEIGHT_CODES[instance.weight] == WEIGHT_CODES[weight]
            ):
                instance.weight = weight
        except KeyError:
            # FIXME: what now
            pass

        try:
            if not instance.width:
                instance.width = ufo_instance.lib[WIDTH_KEY]
        except KeyError:
            # FIXME: what now
            pass

        if ufo_instance.familyName is not None:
            if ufo_instance.familyName != self.font.familyName:
                instance.familyName = ufo_instance.familyName

        smfn = ufo_instance.styleMapFamilyName
        if smfn is not None:
            if smfn.startswith(ufo_instance.familyName):
                smfn = smfn[len(ufo_instance.familyName) :].strip()
            instance.linkStyle = smfn

        if ufo_instance.styleMapStyleName is not None:
            style = ufo_instance.styleMapStyleName
            instance.isBold = "bold" in style
            instance.isItalic = "italic" in style

        if ufo_instance.postScriptFontName is not None:
            instance.fontName = ufo_instance.postScriptFontName

        try:
            instance.manualInterpolation = ufo_instance.lib[MANUAL_INTERPOLATION_KEY]
        except KeyError:
            pass

        try:
            instance.instanceInterpolations = ufo_instance.lib[
                INSTANCE_INTERPOLATIONS_KEY
            ]
        except KeyError:
            # TODO: (jany) compute instanceInterpolations from the location
            # if instance.manualInterpolation: warn about data loss
            pass

        if CUSTOM_PARAMETERS_KEY in ufo_instance.lib:
            for name, value in ufo_instance.lib[CUSTOM_PARAMETERS_KEY]:
                instance.customParameters.append(GSCustomParameter(name, value))

        if ufo_instance.filename and self.minimize_ufo_diffs:
            instance.customParameters[UFO_FILENAME_CUSTOM_PARAM] = ufo_instance.filename

        # FIXME: (jany) cannot `.append()` because no proxy => no parent
        self.font.instances = self.font.instances + [instance]


class InstanceDescriptorAsGSInstance:
    """Wraps a designspace InstanceDescriptor and makes it behave like a
    GSInstance, just enough to use the descriptor as a source of custom
    parameters for `to_ufo_custom_parameters`
    """

    def __init__(self, descriptor):
        self._descriptor = descriptor

        self.customParameters = CustomParametersProxy(None)
        if CUSTOM_PARAMETERS_KEY in descriptor.lib:
            for name, value in descriptor.lib[CUSTOM_PARAMETERS_KEY]:
                self.customParameters[name] = value
        self.properties = PropertiesProxy(None)
        if PROPERTIES_KEY in descriptor.lib:
            for name, value in descriptor.lib[PROPERTIES_KEY]:
                self.properties[name] = value


def _set_class_from_instance(ufo, designspace, instance, axis_tag):
    # FIXME: (jany) copy-pasted from above, factor into method?
    assert axis_tag in ("wght", "wdth")

    factory = AxisDefinitionFactory()
    for axis in designspace.axes:
        if axis.tag == axis_tag:
            axis_def = factory.get(axis.tag, axis.name)
            mapping = axis.map
            break
    else:
        # axis not found, try use the default axis definition
        axis_def = WEIGHT_AXIS_DEF if axis_tag == "wght" else WIDTH_AXIS_DEF
        mapping = []

    try:
        design_loc = instance.location[axis_def.name]
    except KeyError:
        user_loc = axis_def.default_user_loc
    else:
        if mapping:
            # Retrieve the user location (weightClass/widthClass)
            # by going through the axis mapping in reverse.
            reverse_mapping = {dl: ul for ul, dl in mapping}
            user_loc = piecewiseLinearMap(design_loc, reverse_mapping)
        else:
            # no mapping means user space location is same as design space
            user_loc = design_loc
    axis_def.set_ufo_user_loc(ufo, user_loc)


def set_weight_class(ufo, designspace, instance):
    """Set ufo.info.openTypeOS2WeightClass according to the user location
    of the designspace instance, as calculated from the axis mapping.
    """
    _set_class_from_instance(ufo, designspace, instance, "wght")


def set_width_class(ufo, designspace, instance):
    """Set ufo.info.openTypeOS2WidthClass according to the user location
    of the designspace instance, as calculated from the axis mapping.
    """
    _set_class_from_instance(ufo, designspace, instance, "wdth")


def apply_instance_data(designspace, include_filenames=None, Font=None):
    """Open UFO instances referenced by designspace, apply Glyphs instance
    data if present, re-save UFOs and return updated UFO Font objects.

    Args:
        designspace: DesignSpaceDocument object or path (str or PathLike) to
            a designspace file.
        include_filenames: optional set of instance filenames (relative to
            the designspace path) to be included. By default all instaces are
            processed.
        Font: a callable(path: str) -> Font, used to load a UFO, such as
            defcon.Font class (default: ufoLib2.Font.open).
    Returns:
        List of opened and updated instance UFOs.
    """
    from fontTools.designspaceLib import DesignSpaceDocument
    from os.path import normcase, normpath

    if Font is None:
        import ufoLib2

        Font = ufoLib2.Font.open

    if hasattr(designspace, "__fspath__"):
        designspace = designspace.__fspath__()
    if isinstance(designspace, str):
        designspace = DesignSpaceDocument.fromfile(designspace)

    basedir = os.path.dirname(designspace.path)
    instance_ufos = []
    if include_filenames is not None:
        include_filenames = {normcase(normpath(p)) for p in include_filenames}

    for designspace_instance in designspace.instances:
        fname = designspace_instance.filename
        assert fname is not None, "instance %r missing required filename" % getattr(
            designspace_instance, "name", designspace_instance
        )
        if include_filenames is not None:
            fname = normcase(normpath(fname))
            if fname not in include_filenames:
                continue

        logger.debug("Applying instance data to %s", fname)
        # fontmake <= 1.4.0 compares the ufo paths returned from this function
        # to the keys of a dict of designspace locations that have been passed
        # through normpath (but not normcase). We do the same.
        ufo = Font(normpath(os.path.join(basedir, fname)))

        apply_instance_data_to_ufo(ufo, designspace_instance, designspace)

        ufo.save()
        instance_ufos.append(ufo)
    return instance_ufos


def apply_instance_data_to_ufo(ufo, instance, designspace):
    """Apply Glyphs instance data to UFO object.

    Args:
        ufo: a defcon-like font object.
        instance: a fontTools.designspaceLib.InstanceDescriptor.
        designspace: a fontTools.designspaceLib.DesignSpaceDocument.
    Returns:
        None.
    """
    if any(axis.tag == "wght" for axis in designspace.axes):
        set_weight_class(ufo, designspace, instance)
    if any(axis.tag == "wdth" for axis in designspace.axes):
        set_width_class(ufo, designspace, instance)

    glyphs_instance = InstanceDescriptorAsGSInstance(instance)
    to_ufo_custom_params(None, ufo, glyphs_instance)



import glyphsLib


def test_glyphs_font_without_propagated_anchors(datadir):
    font = glyphsLib.GSFont(str(datadir.join("AnchorPropagation.glyphs")))

    # These two layers are supposed to be devoid of anchors in the .glyphs source
    # with Glyphs.app expecting to propagate them automatically upon font generation
    assert len(font.glyphs["lam_alefHamzaabove-ar"].layers[0].anchors) == 0
    assert len(font.glyphs["shadda_fatha-ar"].layers[0].anchors) == 0

    # check for `top_1` anchor of lam_alef-ar.short to be in its original location
    layer = font.glyphs["lam_alef-ar.short"].layers[0]
    assert "top_1" in [x.name for x in layer.anchors]
    assert layer.anchors["top_1"].position.x == 498
    assert layer.anchors["top_1"].position.y == 760

    # check for `top_2` anchor of lam_alef-ar.short to be in its original location
    layer = font.glyphs["lam_alef-ar.short"].layers[0]
    assert "top_2" in [x.name for x in layer.anchors]
    assert layer.anchors["top_2"].position.x == 130
    assert layer.anchors["top_2"].position.y == 628


def test_ufo_with_propagated_anchors(datadir):
    ufo = glyphsLib.load_to_ufos(datadir.join("AnchorPropagation.glyphs"))[0]

    # In UFO, the same two glyphs (see above) are supposed to show anchors
    assert len(ufo["lam_alefHamzaabove-ar"].anchors) > 0
    assert len(ufo["shadda_fatha-ar"].anchors) > 0

    # Additionally, we’ll check the anchor positions and compare them with Glyphs.app’s
    # own results (hard coded) from `layer.anchorsTraversingComponents()`

    # In case of `lam_alefHamzaabove-ar` it’s crucial that the `top_2` anchor moved
    # upwards following the `hamza-ar` mark as part of the ligature base glyph

    # lam_alefHamzaabove-ar
    assert "top_2" in [x.name for x in ufo["lam_alefHamzaabove-ar"].anchors]
    for anchor in ufo["lam_alefHamzaabove-ar"].anchors:
        if anchor.name == "top_2":
            assert anchor.x == 129
            assert anchor.y == 950

    # lamHamzaabove_alefHamzaabove-ar (fictional glyph with 2x same component)
    assert "top_1" in [x.name for x in ufo["lamHamzaabove_alefHamzaabove-ar"].anchors]
    assert "top_2" in [x.name for x in ufo["lamHamzaabove_alefHamzaabove-ar"].anchors]
    for anchor in ufo["lamHamzaabove_alefHamzaabove-ar"].anchors:
        if anchor.name == "top_1":
            assert anchor.x == 497
            assert anchor.y == 1082
    for anchor in ufo["lamHamzaabove_alefHamzaabove-ar"].anchors:
        if anchor.name == "top_2":
            assert anchor.x == 129
            assert anchor.y == 950

    # shadda_fatha-ar
    assert "top" in [x.name for x in ufo["shadda_fatha-ar"].anchors]
    for anchor in ufo["shadda_fatha-ar"].anchors:
        if anchor.name == "top":
            assert anchor.x == 160
            assert anchor.y == 971



import re
import fnmatch
from collections import OrderedDict


def _like(got, expected):
    # LIKE is similar to Unix shell-style wildcards supported by fnmatch
    return fnmatch.fnmatchcase(str(got), expected)


def _contains(got, expected):
    try:
        return expected in got
    except TypeError:
        return expected in str(got)


class TokenExpander:
    number_token_re = r"\$\{([^}]+)\}"
    glyph_predicate_re = r"\$\[([^\]]+)\]"
    glyph_property_re = r"\$\{([^}]+:[^}]+)\}"
    bare_number_value_re = r"\$(\w+)\b"

    def __init__(self, font, master):
        self.font = font
        self.master = master

    def expand(self, featurecode):
        self.featurecode = featurecode
        self.output = ""
        self.position = 0
        while len(self.featurecode) > 0:
            if self.featurecode[0] == "$":
                self.parse_token()
            m = re.match(r"^[^$]+", self.featurecode)
            if m:
                self.output += m[0]
                self._consume_match(m)
        return self.output

    def _consume_match(self, m):
        self.position += len(m[0])
        self.featurecode = self.featurecode[len(m[0]) :]

    def parse_token(self):
        for regexp, parser in [
            (self.glyph_property_re, self.parse_glyph_property),
            (self.number_token_re, self.parse_number_token),
            (self.glyph_predicate_re, self.parse_glyph_predicate),
            (self.bare_number_value_re, self.parse_bare_number_value),
        ]:
            m = re.match(regexp, self.featurecode)
            if m:
                self.output += parser(m[1])
                self._consume_match(m)
                return

        raise ValueError(
            "Unknown token type: '%s' at position %i"
            % (self.featurecode[:10], self.position)
        )

    def parse_bare_number_value(self, number):
        # Find index of number
        try:
            index = [metric.name for metric in self.font.numbers].index(number)
        except ValueError as e:
            raise ValueError(
                "Unknown number token '$%s' at position %i" % (number, self.position)
            ) from e

        value = self.master.numbers[index]
        # We don't add this to the output, because we use the same routine in
        # number tokens
        return str(value)

    def parse_number_token(self, token, layer=None):
        # These things can contain: number tokens, literal numbers, +-*/, space
        # We will also allow ()
        expression = ""
        originaltoken = token
        while len(token) > 0:
            if token[0] in "0123456789.+-*/() ":
                expression += token[0]
                token = token[1:]
                continue
            # Assume it's a property or a number token
            if layer is not None:
                m = re.match(r"^([a-zA-Z0-9_.]+)", token)
            else:
                m = re.match(r"^(\w+)", token)
            if not m:
                raise ValueError(
                    "Unknown character %s in number token '%s' at position %i"
                    % (token[0], originaltoken, self.position)
                )
            if layer is not None:
                try:
                    expression += self._get_value_for_layer(layer, m[1])
                except ValueError:
                    expression += self.parse_bare_number_value(m[1])
            else:
                expression += self.parse_bare_number_value(m[1])
            token = token[len(m[0]) :]
        # This expression is now just numbers and operators - safe to eval,
        # but needs to be an integer
        return "%i" % eval(expression)

    def parse_glyph_property(self, token):
        glyphname, token = token.split(":")
        glyph = self.font.glyphs[glyphname]
        layer = glyph.layers[self.master.id]
        return self.parse_number_token(token, layer)

    def parse_glyph_predicate(self, token):
        self.originaltoken = token
        self.glyph_predicate = token  # Use a little subparser thing here
        return " ".join(self._parse_glyph_predicate_to_array())

    # All following methods are part of the glyph predicate subparser

    gsglyph_predicate_objects = [
        "hasAlignedWidth",
        "hasAnnotations",
        "hasComponents",
        "hasCorners",
        "hasCustomGlyphInfo",
        "hasHints",
        "hasPostScriptHints",
        "hasSpecialLayers",
        "hasTrueTypeHints",
        "isAligned",
        "isAppleColorGlyph",
        "isAnyColorGlyph",
        "isFullColorGlyph",
        "isCornerGlyph",
        "isHangulKeyGlyph",
        "isSVGColorGlyph",
        "isColorPaletteGlyph",
        "isSmartGlyph",
        "justLocked",
        "locked",
        "mastersCompatible",
        "outlineHasChanges",
        "case",
        "changeCount",
        "colorIndex",
        "countOfLayers",
        "countOfPartsSettings",
        "countOfTags",
        "countOfUnicodes",
        "direction",
        "baseString",
        "bottomKerningGroup",
        "bottomMetricsKey",
        "category",
        "charName",
        "charString",
        "description",
        "glyphDataEntryString",
        "lastChange",
        "leftKerningGroup",
        "leftKerningKey",
        "leftMetricsKey",
        "name",
        "note",
        "production",
        "productionName",
        "rightKerningGroup",
        "rightKerningKey",
        "rightMetricsKey",
        "script",
        "sortName",
        "sortNameKeep",
        "string",
        "subCategory",
        "tags",
        "topKerningGroup",
        "topMetricsKey",
        "unicode",
        "unicodeChar",
        "unicodeString",
        "vertWidthMetricsKey",
        "widthMetricsKey",
    ]
    gsglyph_predicate_object_re = (
        r"^\s*(" + "|".join(gsglyph_predicate_objects) + r"\b)"
    )
    comparators_re = (
        r"(?i)\s*(beginswith|contains|endswith|like|matches|==?|>=|"
        r"=>|<=|=<|!=|<>|>|<|between|in)\s*"
    )

    def _parse_glyph_predicate_to_array(self):
        invert = self._parse_optional_not()
        # Handle 'value in property' by flipping to 'property contains value'
        if re.match(self.gsglyph_predicate_object_re, self.glyph_predicate):
            obj = self._parse_object()
            comparator = self._parse_comparator()
            if comparator in ("in", "between"):
                expected = self._parse_aggregate()
            else:
                expected = self._parse_value()
        elif re.search(r"(?i)\bin\b", self.glyph_predicate):
            expected = self._parse_value()
            comparator = self._parse_comparator()
            assert comparator == "in"
            obj = self._parse_object()
            comparator = "contains"
        else:
            self._parse_object()  # will raise with proper error

        glyphs = OrderedDict({})
        for g in self.font.glyphs:
            if not g.export:
                continue
            v = self._get_value_for_glyph(g, obj)
            truth = self._compare(v, comparator, expected)
            if invert:
                truth = not truth
            if truth:
                glyphs[g.name] = None

        compound = self._parse_compound()
        if compound:
            if compound == "OR":
                for g2 in self._parse_glyph_predicate_to_array():
                    glyphs[g2] = None
            else:  # AND
                otherset = set(self._parse_glyph_predicate_to_array())
                return [g for g in glyphs.keys() if g in otherset]

        return list(glyphs.keys())

    def _parse_optional_not(self):
        m = re.match(r"(?i)^\s*(not|!)\s+", self.glyph_predicate)
        if m:
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            return True
        return False

    def _parse_error_in_predicate(self, thing):
        found = re.match(r"^\s*(\S+)", self.glyph_predicate)
        raise ValueError(
            "Expected a %s, found '%s' in predicate '%s'"
            % (thing, found[0], self.originaltoken)
        )

    def _parse_object(self):
        m = re.match(
            self.gsglyph_predicate_object_re, self.glyph_predicate
        ) or self._parse_error_in_predicate("known glyph predicate object")
        self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
        return m[1]

    def _parse_comparator(self):
        m = re.match(
            self.comparators_re, self.glyph_predicate
        ) or self._parse_error_in_predicate("comparator")
        normalize_comparators = {"=": "==", "=>": ">=", "=<": "<=", "<>": "!="}
        self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
        return normalize_comparators.get(m[1], m[1]).lower()

    def _parse_aggregate(self):
        m = re.match(r"\s*\{", self.glyph_predicate)
        if not m:
            self._parse_error_in_predicate("aggregate '{...}'")
        self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
        values = [self._parse_value()]
        while m := re.match(r"\s*,", self.glyph_predicate):
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            values.append(self._parse_value())
        m = re.match(r"\s*}", self.glyph_predicate)
        if not m:
            self._parse_error_in_predicate("'}'")
        self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
        return values

    def _parse_value(self):
        m = re.match(r'\s*"([^"]+)"', self.glyph_predicate) or re.match(
            r"\s*'([^']+)'", self.glyph_predicate
        )  # Doesn't handle string escapes
        if m:
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            return m[1]
        m = re.match(r"(?i)\s*(yes|true)", self.glyph_predicate)
        if m:
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            return True
        m = re.match(r"(?i)\s*(no|false)", self.glyph_predicate)
        if m:
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            return False
        # The tutorial says number values can be floats, but I don't think they can
        m = re.match(r"\s*(\d+)", self.glyph_predicate)
        if m:
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            return int(m[1])

        # Keyword constants
        m = re.match(r"\s*(\w+)", self.glyph_predicate)
        if m:
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            return m[1]
        self._parse_error_in_predicate("value")

    def _get_value_for_anchor(self, layer, value):
        _, anchorname, attr = value.split(".")
        anchor = layer.anchors[anchorname]
        if anchor is None:
            raise ValueError(
                "Unknown anchor '%s' at position %i" % (anchorname, self.position)
            )
        try:
            return str(getattr(anchor.position, attr))
        except AttributeError as e:
            raise ValueError(
                "Unknown anchor attribute '%s' at position %i" % (attr, self.position)
            ) from e

    def _get_value_for_layer(self, layer, value):
        if value.startswith("anchors."):
            return self._get_value_for_anchor(layer, value)

        try:
            return str(getattr(layer, value))
        except AttributeError as e:
            raise ValueError(
                "Unknown glyph property '%s' at position %i" % (value, self.position)
            ) from e

    def _get_value_for_glyph(self, g, value):
        try:
            attrib = getattr(g, value)
            if callable(attrib):
                return attrib()
            else:
                return attrib
        except AttributeError as exc:
            raise ValueError(
                "Glyphs attribute %s used in predicate '%s'"
                " but glyphsLib does not support it" % (value, self.originaltoken)
            ) from exc

    apply_comparators = {
        "beginswith": lambda got, exp: str(got).startswith(exp),
        "contains": _contains,
        "endswith": lambda got, exp: str(got).endswith(exp),
        "like": _like,
        "matches": lambda got, exp: re.search(exp, str(got)),
        "==": lambda got, exp: got == exp,
        "!=": lambda got, exp: got != exp,
        ">=": lambda got, exp: got >= exp,
        "<=": lambda got, exp: got <= exp,
        ">": lambda got, exp: got > exp,
        "<": lambda got, exp: got < exp,
        "between": lambda got, exp: got >= exp[0] and got <= exp[1],
        "in": lambda got, exp: got in exp,
    }

    def _compare(self, got, comparator, expected):
        return self.apply_comparators[comparator](got, expected)

    def _parse_compound(self):
        m = re.match(r"(?i)\s*(and|&&)", self.glyph_predicate)
        if m:
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            return "AND"

        m = re.match(r"(?i)\s*(or|\|\|)", self.glyph_predicate)
        if m:
            self.glyph_predicate = self.glyph_predicate[len(m[0]) :]
            return "OR"

        return False


class PassThruExpander:
    def expand(self, token):
        return token



#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Goal: have a unit test for each parameter that is mentioned in the UFO spec.
Each test will check that this parameter is round-tripped and, when relevant,
that the Glyphs storage of the value has the correct Glyphs meaning.

http://unifiedfontobject.org/versions/ufo3/fontinfo.plist/
"""

import os
import pathlib
import pytest
from collections import namedtuple

from glyphsLib import to_glyphs, to_ufos, classes


def section(name, *fields):
    return pytest.param(fields, id=name)


def skip_section(name, *fields):
    return pytest.param(fields, id=name, marks=pytest.mark.skip)


Field = namedtuple("Field", "name test_value")


ufo_info_spec = [
    section(
        "Generic Identification Information",
        # familyName	string	Family name.
        Field("familyName", "Ronoto Sans"),
        # styleName	string	Style name.
        Field("styleName", "Condensed"),
        # styleMapFamilyName string Family name used for bold, italic and bold
        # italic style mapping.
        Field("styleMapFamilyName", "Ronoto Sans Condensed"),
        # styleMapStyleName string Style map style. The possible values are
        # regular, italic, bold and bold italic. These are case sensitive.
        Field("styleMapStyleName", "regular"),
        # versionMajor integer Major version.
        Field("versionMajor", 1),
        # versionMinor	non-negative integer	Minor version.
        Field("versionMinor", 12),
        # year integer The year the font was created. This attribute is
        # deprecated as of version 2. Its presence should not be relied upon by
        # authoring tools. However, it may occur in a font's info so authoring
        # tools should preserve it if present.
        Field("year", 2013),
    ),
    section(
        "Generic Legal Information",
        # copyright	string	Copyright statement.
        Field("copyright", "© Glooble"),
        # trademark	string	Trademark statement.
        Field("trademark", "Ronoto™® is a trademark of Glooble Inc. Ltd."),
    ),
    section(
        "Generic Dimension Information",
        # unitsPerEm	non-negative integer or float	Units per em.
        Field("unitsPerEm", 1234),
        # descender	integer or float	Descender value.
        Field("descender", 123.7),
        # xHeight	integer or float	x-height value.
        Field("xHeight", 456),
        # capHeight	integer or float	Cap height value.
        Field("capHeight", 789),
        # ascender	integer or float	Ascender value.
        Field("ascender", 789.1),
        # italicAngle integer or float Italic angle. This must be an angle in
        # counter-clockwise degrees from the vertical.
        Field("italicAngle", -12.5),
    ),
    section(
        "Generic Miscellaneous Information",
        # note	string	Arbitrary note about the font.
        Field("note", "Bla bla"),
    ),
    section(
        "OpenType head Table Fields",
        # openTypeHeadCreated string Creation date. Expressed as a string of
        # the format “YYYY/MM/DD HH:MM:SS”. “YYYY/MM/DD” is year/month/day. The
        # month must be in the range 1-12 and the day must be in the range
        # 1-end of month. “HH:MM:SS” is hour:minute:second. The hour must be in
        # the range 0:23. The minute and second must each be in the range 0-59.
        Field("openTypeHeadCreated", "2014/02/28 19:20:48"),
        # openTypeHeadLowestRecPPEM non-negative integer Smallest readable size
        # in pixels. Corresponds to the OpenType head table lowestRecPPEM
        # field.
        Field("openTypeHeadLowestRecPPEM", 12),
        # openTypeHeadFlags list A list of bit numbers indicating the flags.
        # The bit numbers are listed in the OpenType head specification.
        # Corresponds to the OpenType head table flags field.
        Field("openTypeHeadFlags", [2, 3, 11]),
    ),
    section(
        "OpenType hhea Table Fields",
        # openTypeHheaAscender integer Ascender value. Corresponds to the
        # OpenType hhea table Ascender field.
        Field("openTypeHheaAscender", 123),
        # openTypeHheaDescender integer Descender value. Corresponds to the
        # OpenType hhea table Descender field.
        Field("openTypeHheaDescender", 456),
        # openTypeHheaLineGap integer Line gap value. Corresponds to the
        # OpenType hhea table LineGap field.
        Field("openTypeHheaLineGap", 789),
        # openTypeHheaCaretSlopeRise integer Caret slope rise value.
        # Corresponds to the OpenType hhea table caretSlopeRise field.
        Field("openTypeHheaCaretSlopeRise", 800),
        # openTypeHheaCaretSlopeRun integer Caret slope run value. Corresponds
        # to the OpenType hhea table caretSlopeRun field.
        Field("openTypeHheaCaretSlopeRun", 100),
        # openTypeHheaCaretOffset integer Caret offset value. Corresponds to
        # the OpenType hhea table caretOffset field.
        Field("openTypeHheaCaretOffset", 20),
    ),
    section(
        "OpenType name Table Fields",
        # openTypeNameDesigner	string	Designer name.
        # Corresponds to the OpenType name table name ID 9.
        Field("openTypeNameDesigner", "Bob"),
        # openTypeNameDesignerURL	string	URL for the designer.
        # Corresponds to the OpenType name table name ID 12.
        Field("openTypeNameDesignerURL", "http://bob.me/"),
        # openTypeNameManufacturer	string	Manufacturer name.
        # Corresponds to the OpenType name table name ID 8.
        Field("openTypeNameManufacturer", "Exemplary Type"),
        # openTypeNameManufacturerURL	string	Manufacturer URL.
        # Corresponds to the OpenType name table name ID 11.
        Field("openTypeNameManufacturerURL", "http://exemplary.type"),
        # openTypeNameLicense	string	License text.
        # Corresponds to the OpenType name table name ID 13.
        Field("openTypeNameLicense", "OFL 1.1"),
        # openTypeNameLicenseURL	string	URL for the license.
        # Corresponds to the OpenType name table name ID 14.
        Field("openTypeNameLicenseURL", "http://scripts.sil.org/OFL"),
        # openTypeNameVersion	string	Version string.
        # Corresponds to the OpenType name table name ID 5.
        Field("openTypeNameVersion", "Version 2.003"),
        # openTypeNameUniqueID	string	Unique ID string.
        # Corresponds to the OpenType name table name ID 3.
        Field("openTypeNameUniqueID", "2.003;Exemplary Sans Bold Large Display"),
        # openTypeNameDescription	string	Description of the font.
        # Corresponds to the OpenType name table name ID 10.
        Field("openTypeNameDescription", "Best used\nfor typesetting\nhaikus"),
        # openTypeNamePreferredFamilyName	string	Preferred family name.
        # Corresponds to the OpenType name table name ID 16.
        Field("openTypeNamePreferredFamilyName", "Exemplary Sans"),
        # openTypeNamePreferredSubfamilyName	string	Preferred subfamily name.
        # Corresponds to the OpenType name table name ID 17.
        Field("openTypeNamePreferredSubfamilyName", "Bold Large Display"),
        # openTypeNameCompatibleFullName	string	Compatible full name.
        # Corresponds to the OpenType name table name ID 18.
        Field("openTypeNameCompatibleFullName", "Exemplary Sans Bold Large Display"),
        # openTypeNameSampleText	string	Sample text.
        # Corresponds to the OpenType name table name ID 19.
        Field("openTypeNameSampleText", "Pickles are our friends"),
        # openTypeNameWWSFamilyName	string	WWS family name.
        # Corresponds to the OpenType name table name ID 21.
        Field("openTypeNameWWSFamilyName", "Exemplary Sans Display"),
        # openTypeNameWWSSubfamilyName	string	WWS Subfamily name.
        # Corresponds to the OpenType name table name ID 22.
        Field("openTypeNameWWSSubfamilyName", "Bold Large"),
    ),
    section(
        "OpenType OS/2 Table Fields",
        # openTypeOS2WidthClass integer Width class value. Must be in the range
        # 1-9 Corresponds to the OpenType OS/2 table usWidthClass field.
        Field("openTypeOS2WidthClass", 7),
        # openTypeOS2WeightClass integer Weight class value. Must be a
        # non-negative integer. Corresponds to the OpenType OS/2 table
        # usWeightClass field.
        Field("openTypeOS2WeightClass", 700),
        # openTypeOS2Selection list A list of bit numbers indicating the bits
        # that should be set in fsSelection. The bit numbers are listed in the
        # OpenType OS/2 specification.
        # Corresponds to the OpenType OS/2 table selection field.
        # Note: Bits 0 (italic), 5 (bold) and 6 (regular) must not be set here.
        # These bits should be taken from the generic styleMapStyle attribute.
        Field("openTypeOS2Selection", [7, 8]),
        # openTypeOS2VendorID string Four character identifier for the creator
        # of the font. Corresponds to the OpenType OS/2 table achVendID field.
        Field("openTypeOS2VendorID", "EXTY"),
        # openTypeOS2Panose list The list must contain 10 non-negative integers
        # that represent the setting for each category in the Panose
        # specification. The integers correspond with the option numbers in
        # each of the Panose categories. This corresponds to the OpenType OS/2
        # table Panose field.
        Field("openTypeOS2Panose", [2, 11, 8, 5, 3, 4, 0, 0, 0, 0]),
        # openTypeOS2FamilyClass list Two integers representing the IBM font
        # class and font subclass of the font. The first number, representing
        # the class ID, must be in the range 0-14. The second number,
        # representing the subclass, must be in the range 0-15. The numbers are
        # listed in the OpenType OS/2 specification.
        # Corresponds to the OpenType OS/2 table sFamilyClass field.
        Field("openTypeOS2FamilyClass", [8, 2]),
        # openTypeOS2UnicodeRanges list A list of bit numbers that are
        # supported Unicode ranges in the font. The bit numbers are listed in
        # the OpenType OS/2 specification. Corresponds to the OpenType OS/2
        # table ulUnicodeRange1, ulUnicodeRange2, ulUnicodeRange3 and
        # ulUnicodeRange4 fields.
        Field("openTypeOS2UnicodeRanges", [0, 1, 2, 3, 37, 79, 122]),
        # openTypeOS2CodePageRanges list A list of bit numbers that are
        # supported code page ranges in the font. The bit numbers are listed in
        # the OpenType OS/2 specification. Corresponds to the OpenType OS/2
        # table ulCodePageRange1 and ulCodePageRange2 fields.
        Field("openTypeOS2CodePageRanges", [0, 1, 29, 58]),
        # openTypeOS2TypoAscender	integer	Ascender value.
        # Corresponds to the OpenType OS/2 table sTypoAscender field.
        Field("openTypeOS2TypoAscender", 1000),
        # openTypeOS2TypoDescender	integer	Descender value.
        # Corresponds to the OpenType OS/2 table sTypoDescender field.
        Field("openTypeOS2TypoDescender", -234),
        # openTypeOS2TypoLineGap	integer	Line gap value.
        # Corresponds to the OpenType OS/2 table sTypoLineGap field.
        Field("openTypeOS2TypoLineGap", 456),
        # openTypeOS2WinAscent	non-negative integer	Ascender value.
        # Corresponds to the OpenType OS/2 table usWinAscent field.
        Field("openTypeOS2WinAscent", 1500),
        # openTypeOS2WinDescent	non-negative integer	Descender value.
        # Corresponds to the OpenType OS/2 table usWinDescent field.
        Field("openTypeOS2WinDescent", 750),
        # openTypeOS2Type list A list of bit numbers indicating the embedding
        # type. The bit numbers are listed in the OpenType OS/2 specification.
        # Corresponds to the OpenType OS/2 table fsType field.
        Field("openTypeOS2Type", [3, 8]),
        # openTypeOS2SubscriptXSize	integer	Subscript horizontal font size.
        # Corresponds to the OpenType OS/2 table ySubscriptXSize field.
        Field("openTypeOS2SubscriptXSize", 123),
        # openTypeOS2SubscriptYSize	integer	Subscript vertical font size.
        # Corresponds to the OpenType OS/2 table ySubscriptYSize field.
        Field("openTypeOS2SubscriptYSize", 246),
        # openTypeOS2SubscriptXOffset	integer	Subscript x offset.
        # Corresponds to the OpenType OS/2 table ySubscriptXOffset field.
        Field("openTypeOS2SubscriptXOffset", -6),
        # openTypeOS2SubscriptYOffset	integer	Subscript y offset.
        # Corresponds to the OpenType OS/2 table ySubscriptYOffset field.
        Field("openTypeOS2SubscriptYOffset", 100),
        # openTypeOS2SuperscriptXSize integer Superscript horizontal font size.
        # Corresponds to the OpenType OS/2 table ySuperscriptXSize field.
        Field("openTypeOS2SuperscriptXSize", 124),
        # openTypeOS2SuperscriptYSize integer Superscript vertical font size.
        # Corresponds to the OpenType OS/2 table ySuperscriptYSize field.
        Field("openTypeOS2SuperscriptYSize", 248),
        # openTypeOS2SuperscriptXOffset	integer	Superscript x offset.
        # Corresponds to the OpenType OS/2 table ySuperscriptXOffset field.
        Field("openTypeOS2SuperscriptXOffset", -8),
        # openTypeOS2SuperscriptYOffset	integer	Superscript y offset.
        # Corresponds to the OpenType OS/2 table ySuperscriptYOffset field.
        Field("openTypeOS2SuperscriptYOffset", 400),
        # openTypeOS2StrikeoutSize	integer	Strikeout size.
        # Corresponds to the OpenType OS/2 table yStrikeoutSize field.
        Field("openTypeOS2StrikeoutSize", 56),
        # openTypeOS2StrikeoutPosition	integer	Strikeout position.
        # Corresponds to the OpenType OS/2 table yStrikeoutPosition field.
        Field("openTypeOS2StrikeoutPosition", 200),
    ),
    section(
        "OpenType vhea Table Fields",
        # openTypeVheaVertTypoAscender integer Ascender value. Corresponds to
        # the OpenType vhea table vertTypoAscender field.
        Field("openTypeVheaVertTypoAscender", 123),
        # openTypeVheaVertTypoDescender integer Descender value. Corresponds to
        # the OpenType vhea table vertTypoDescender field.
        Field("openTypeVheaVertTypoDescender", 456),
        # openTypeVheaVertTypoLineGap integer Line gap value. Corresponds to
        # the OpenType vhea table vertTypoLineGap field.
        Field("openTypeVheaVertTypoLineGap", 789),
        # openTypeVheaCaretSlopeRise integer Caret slope rise value.
        # Corresponds to the OpenType vhea table caretSlopeRise field.
        Field("openTypeVheaCaretSlopeRise", 23),
        # openTypeVheaCaretSlopeRun integer Caret slope run value. Corresponds
        # to the OpenType vhea table caretSlopeRun field.
        Field("openTypeVheaCaretSlopeRun", 1024),
        # openTypeVheaCaretOffset integer Caret offset value. Corresponds to
        # the OpenType vhea table caretOffset field.
        Field("openTypeVheaCaretOffset", 50),
    ),
    section(
        "PostScript Specific Data",
        # postscriptFontName string Name to be used for the FontName field in
        # Type 1/CFF table.
        Field("postscriptFontName", "Exemplary-Sans"),
        # postscriptFullName string Name to be used for the FullName field in
        # Type 1/CFF table.
        Field("postscriptFullName", "Exemplary Sans Bold Whatever"),
        # postscriptSlantAngle integer or float Artificial slant angle. This
        # must be an angle in counter-clockwise degrees from the vertical.
        Field("postscriptSlantAngle", -15.5),
        # postscriptUniqueID integer A unique ID number as defined in the Type
        # 1/CFF specification.
        Field("postscriptUniqueID", 123456789),
        # postscriptUnderlineThickness integer or float Underline thickness
        # value. Corresponds to the Type 1/CFF/post table UnderlineThickness
        # field.
        Field("postscriptUnderlineThickness", 70),
        # postscriptUnderlinePosition integer or float Underline position
        # value. Corresponds to the Type 1/CFF/post table UnderlinePosition
        # field.
        Field("postscriptUnderlinePosition", -50),
        # postscriptIsFixedPitch boolean Indicates if the font is monospaced.
        # An authoring tool could calculate this automatically, but the
        # designer may wish to override this setting. This corresponds to the
        # Type 1/CFF isFixedPitched field
        Field("postscriptIsFixedPitch", True),
        # postscriptBlueValues list A list of up to 14 integers or floats
        # specifying the values that should be in the Type 1/CFF BlueValues
        # field. This list must contain an even number of integers following
        # the rules defined in the Type 1/CFF specification.
        Field("postscriptBlueValues", [-200, -185, -15, 0, 500, 515]),
        # postscriptOtherBlues list A list of up to 10 integers or floats
        # specifying the values that should be in the Type 1/CFF OtherBlues
        # field. This list must contain an even number of integers following
        # the rules defined in the Type 1/CFF specification.
        Field("postscriptOtherBlues", [-315, -300, 385, 400]),
        # postscriptFamilyBlues list A list of up to 14 integers or floats
        # specifying the values that should be in the Type 1/CFF FamilyBlues
        # field. This list must contain an even number of integers following
        # the rules defined in the Type 1/CFF specification.
        Field("postscriptFamilyBlues", [-210, -195, -25, 0, 510, 525]),
        # postscriptFamilyOtherBlues list A list of up to 10 integers or floats
        # specifying the values that should be in the Type 1/CFF
        # FamilyOtherBlues field. This list must contain an even number of
        # integers following the rules defined in the Type 1/CFF specification.
        Field("postscriptFamilyOtherBlues", [-335, -330, 365, 480]),
        # postscriptStemSnapH list List of horizontal stems sorted in the order
        # specified in the Type 1/CFF specification. Up to 12 integers or
        # floats are possible. This corresponds to the Type 1/CFF StemSnapH
        # field.
        Field("postscriptStemSnapH", [-10, 40, 400, 789]),
        # postscriptStemSnapV list List of vertical stems sorted in the order
        # specified in the Type 1/CFF specification. Up to 12 integers or
        # floats are possible. This corresponds to the Type 1/CFF StemSnapV
        # field.
        Field("postscriptStemSnapV", [-500, -40, 0, 390, 789]),
        # postscriptBlueFuzz integer or float BlueFuzz value. This corresponds
        # to the Type 1/CFF BlueFuzz field.
        Field("postscriptBlueFuzz", 2),
        # postscriptBlueShift integer or float BlueShift value. This
        # corresponds to the Type 1/CFF BlueShift field.
        Field("postscriptBlueShift", 10),
        # postscriptBlueScale float BlueScale value. This corresponds to the
        # Type 1/CFF BlueScale field.
        Field("postscriptBlueScale", 0.0256),
        # postscriptForceBold boolean Indicates how the Type 1/CFF ForceBold
        # field should be set.
        Field("postscriptForceBold", True),
        # postscriptDefaultWidthX integer or float Default width for glyphs.
        Field("postscriptDefaultWidthX", 250),
        # postscriptNominalWidthX integer or float Nominal width for glyphs.
        Field("postscriptNominalWidthX", 10),
        # postscriptWeightName string A string indicating the overall weight of
        # the font. This corresponds to the Type 1/CFF Weight field. It should
        # be in sync with the openTypeOS2WeightClass value.
        Field("postscriptWeightName", "Bold"),
        # postscriptDefaultCharacter string The name of the glyph that should
        # be used as the default character in PFM files.
        Field("postscriptDefaultCharacter", "a"),
        # postscriptWindowsCharacterSet integer The Windows character set. The
        # values are defined below.
        Field("postscriptWindowsCharacterSet", 4),
    ),
    section(
        "Macintosh FOND Resource Data",
        # macintoshFONDFamilyID integer Family ID number. Corresponds to the
        # ffFamID in the FOND resource.
        Field("macintoshFONDFamilyID", 12345),
        # macintoshFONDName string Font name for the FOND resource.
        Field("macintoshFONDName", "ExemplarySansBold"),
    ),
    skip_section(
        "WOFF Data",
        # woffMajorVersion	non-negative integer	Major version of the font.
        Field("woffMajorVersion", 1),
        # woffMinorVersion	non-negative integer	Minor version of the font.
        Field("woffMinorVersion", 12),
        # woffMetadataUniqueID dictionary Identification string. Corresponds to
        # the WOFF uniqueid. The dictionary must follow the WOFF Metadata
        # Unique ID Record structure.
        Field("woffMetadataUniqueID", {}),
        # woffMetadataVendor dictionary Font vendor. Corresponds to the WOFF
        # vendor element. The dictionary must follow the the WOFF Metadata
        # Vendor Record structure.
        Field("woffMetadataVendor", {}),
        # woffMetadataCredits dictionary Font credits. Corresponds to the WOFF
        # credits element. The dictionary must follow the WOFF Metadata Credits
        # Record structure.
        Field("woffMetadataCredits", {}),
        # woffMetadataDescription dictionary Font description. Corresponds to
        # the WOFF description element. The dictionary must follow the WOFF
        # Metadata Description Record structure.
        Field("woffMetadataDescription", {}),
        # woffMetadataLicense dictionary Font description. Corresponds to the
        # WOFF license element. The dictionary must follow the WOFF Metadata
        # License Record structure.
        Field("woffMetadataLicense", {}),
        # woffMetadataCopyright dictionary Font copyright. Corresponds to the
        # WOFF copyright element. The dictionary must follow the WOFF Metadata
        # Copyright Record structure.
        Field("woffMetadataCopyright", {}),
        # woffMetadataTrademark dictionary Font trademark. Corresponds to the
        # WOFF trademark element. The dictionary must follow the WOFF Metadata
        # Trademark Record structure.
        Field("woffMetadataTrademark", {}),
        # woffMetadataLicensee dictionary Font licensee. Corresponds to the
        # WOFF licensee element. The dictionary must follow the WOFF Metadata
        # Licensee Record structure.
        Field("woffMetadataLicensee", {}),
        # woffMetadataExtensions list List of metadata extension records. The
        # dictionaries must follow the WOFF Metadata Extension Record
        # structure. There must be at least one extension record in the list.
        Field("woffMetadataExtensions", {}),
    ),
]


@pytest.mark.parametrize("fields", ufo_info_spec)
def test_info(fields, tmpdir, ufo_module):
    ufo = ufo_module.Font()

    for field in fields:
        setattr(ufo.info, field.name, field.test_value)

    font = to_glyphs([ufo], minimize_ufo_diffs=True)
    filename = os.path.join(str(tmpdir), "font.glyphs")
    font.save(filename)
    font = classes.GSFont(filename)
    (ufo,) = to_ufos(font, ufo_module=ufo_module)

    for field in fields:
        assert getattr(ufo.info, field.name) == field.test_value


def test_info_gasp_ranges(tmpdir, ufo_module):
    ufo = ufo_module.Font()

    gasp_ranges = [
        {"rangeMaxPPEM": 16, "rangeGaspBehavior": [0]},
        {"rangeMaxPPEM": 65535, "rangeGaspBehavior": [0, 1]},
    ]
    ufo.info.openTypeGaspRangeRecords = gasp_ranges

    font = to_glyphs([ufo], minimize_ufo_diffs=True)
    filename = os.path.join(str(tmpdir), "font.glyphs")
    font.save(filename)
    font = classes.GSFont(filename)
    (ufo,) = to_ufos(font, ufo_module=ufo_module)

    assert [dict(r) for r in ufo.info.openTypeGaspRangeRecords] == gasp_ranges


def test_info_name_records(tmpdir, ufo_module):
    ufo = ufo_module.Font()

    name_records = [
        {
            "nameID": 19,
            "platformID": 1,
            "encodingID": 0,
            "languageID": 1,
            "string": "Les cornichons sont nos amis",
        },
        {
            "nameID": 1,
            "platformID": 3,
            "encodingID": 1,
            "languageID": 0x0410,
            "string": "Illustrativo Sans",
        },
    ]

    ufo.info.openTypeNameRecords = name_records

    font = to_glyphs([ufo], minimize_ufo_diffs=True)
    filename = os.path.join(str(tmpdir), "font.glyphs")
    font.save(filename)
    font = classes.GSFont(filename)
    (ufo,) = to_ufos(font, ufo_module=ufo_module)

    print(ufo.info.openTypeNameRecords)
    assert [dict(r) for r in ufo.info.openTypeNameRecords] == name_records


def test_localized_properties(tmpdir, ufo_module):
    filename = (
        pathlib.Path(__file__).parent.parent / "data" / "GlyphsFileFormatv3.glyphs"
    )
    font = classes.GSFont(filename)

    ufos = to_ufos(font, ufo_module=ufo_module)
    for ufo in ufos:
        assert ufo.info.copyright == "Default Copyright1"
        assert [dict(r) for r in ufo.info.openTypeNameRecords] == [
            {
                "nameID": 0,
                "platformID": 3,
                "encodingID": 1,
                "languageID": 1031,
                "string": "Deutsches Copyright",
            }
        ]



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from collections import OrderedDict

import pytest

from fontTools.designspaceLib import DesignSpaceDocument
from glyphsLib import classes
from glyphsLib.types import BinaryData
from glyphsLib.builder.constants import (
    GLYPHLIB_PREFIX,
    FONT_CUSTOM_PARAM_PREFIX,
    UFO2FT_FEATURE_WRITERS_KEY,
    DEFAULT_FEATURE_WRITERS,
    GLYPHS_MATH_CONSTANTS_KEY,
    GLYPHS_MATH_EXTENDED_SHAPE_KEY,
    GLYPHS_MATH_VARIANTS_KEY,
)
from glyphsLib import to_glyphs, to_ufos, to_designspace


# GOAL: Test the translations between the various UFO lib and Glyphs userData.
# See the associated UML diagram: `lib_and_user_data.png`


def test_designspace_lib_equivalent_to_font_user_data(tmpdir):
    designspace = DesignSpaceDocument()
    designspace.lib["designspaceLibKey1"] = "designspaceLibValue1"

    # Save to disk and reload the designspace to test the write/read of lib
    path = os.path.join(str(tmpdir), "test.designspace")
    designspace.write(path)
    designspace = DesignSpaceDocument()
    designspace.read(path)

    font = to_glyphs(designspace)

    assert font.userData["designspaceLibKey1"] == "designspaceLibValue1"

    designspace = to_designspace(font)

    assert designspace.lib["designspaceLibKey1"] == "designspaceLibValue1"


def test_custom_featureWriters_in_designpace_lib(tmpdir, ufo_module):
    """Test that we can roundtrip custom user-defined ufo2ft featureWriters
    settings that are stored in the designspace lib or GSFont.userData.
    """
    font = classes.GSFont()
    font.masters.append(classes.GSFontMaster())
    kern = classes.GSFeature(name="kern", code="pos a b 100;")
    font.features.append(kern)
    customFeatureWriters = list(DEFAULT_FEATURE_WRITERS) + [
        {"class": "MyCustomWriter", "module": "myCustomWriter"}
    ]
    font.userData[UFO2FT_FEATURE_WRITERS_KEY] = customFeatureWriters

    designspace = to_designspace(font, ufo_module=ufo_module)
    path = str(tmpdir / "test.designspace")
    designspace.write(path)
    for source in designspace.sources:
        source.font.save(str(tmpdir / source.filename))

    designspace2 = DesignSpaceDocument.fromfile(path)

    assert UFO2FT_FEATURE_WRITERS_KEY in designspace2.lib
    assert designspace2.lib[UFO2FT_FEATURE_WRITERS_KEY] == customFeatureWriters

    font2 = to_glyphs(designspace2, ufo_module=ufo_module)

    assert len(font2.userData) == 1
    assert font2.userData[UFO2FT_FEATURE_WRITERS_KEY] == customFeatureWriters


def test_font_user_data_to_ufo_lib():
    # This happens only when not building a designspace
    # Since there is no designspace.lib to store the font userData,
    # the latter is duplicated in each output ufo
    font = classes.GSFont()
    font.masters.append(classes.GSFontMaster())
    font.masters.append(classes.GSFontMaster())
    font.userData["fontUserDataKey"] = "fontUserDataValue"

    ufo1, ufo2 = to_ufos(font)

    assert ufo1.lib[GLYPHLIB_PREFIX + "fontUserData"] == {
        "fontUserDataKey": "fontUserDataValue"
    }
    assert ufo2.lib[GLYPHLIB_PREFIX + "fontUserData"] == {
        "fontUserDataKey": "fontUserDataValue"
    }

    font = to_glyphs([ufo1, ufo2])

    assert font.userData["fontUserDataKey"] == "fontUserDataValue"


def test_DisplayStrings_ufo_lib():
    font = classes.GSFont()
    font.masters.append(classes.GSFontMaster())
    font.masters.append(classes.GSFontMaster())
    font.DisplayStrings = ""

    ufo1, ufo2 = to_ufos(font)
    assert FONT_CUSTOM_PARAM_PREFIX + "DisplayStrings" not in ufo1.lib
    assert FONT_CUSTOM_PARAM_PREFIX + "DisplayStrings" not in ufo2.lib

    font = to_glyphs([ufo1, ufo2])
    assert font.DisplayStrings == ""

    # ---
    font.DisplayStrings = "a"

    ufo1, ufo2 = to_ufos(font)
    assert ufo1.lib[FONT_CUSTOM_PARAM_PREFIX + "DisplayStrings"] == "a"
    assert ufo2.lib[FONT_CUSTOM_PARAM_PREFIX + "DisplayStrings"] == "a"

    font = to_glyphs([ufo1, ufo2])
    assert font.DisplayStrings == "a"


def test_ufo_lib_equivalent_to_font_master_user_data(ufo_module):
    ufo1 = ufo_module.Font()
    ufo1.lib["ufoLibKey1"] = "ufoLibValue1"
    ufo2 = ufo_module.Font()
    ufo2.lib["ufoLibKey2"] = "ufoLibValue2"

    font = to_glyphs([ufo1, ufo2])

    assert font.masters[0].userData["ufoLibKey1"] == "ufoLibValue1"
    assert font.masters[1].userData["ufoLibKey2"] == "ufoLibValue2"

    ufo1, ufo2 = to_ufos(font)

    assert ufo1.lib["ufoLibKey1"] == "ufoLibValue1"
    assert ufo2.lib["ufoLibKey2"] == "ufoLibValue2"
    assert "ufoLibKey2" not in ufo1.lib
    assert "ufoLibKey1" not in ufo2.lib


def test_ufo_data_into_font_master_user_data(tmpdir, ufo_module):
    filename = "org.customTool/ufoData.bin"
    data = b"\x00\x01\xff"
    ufo = ufo_module.Font()
    ufo.data[filename] = data

    font = to_glyphs([ufo])
    # Round-trip to disk for this one because I'm not sure there are other
    # tests that read-write binary data
    path = os.path.join(str(tmpdir), "font.glyphs")
    font.save(path)
    font = classes.GSFont(path)

    # The path in the glyphs file should be os-agnostic (forward slashes)
    assert font.masters[0].userData[GLYPHLIB_PREFIX + "ufoData"] == {
        "org.customTool/ufoData.bin": BinaryData(data)
    }

    (ufo,) = to_ufos(font)

    assert ufo.data[filename] == data


def test_layer_lib_into_master_user_data(ufo_module):
    ufo1 = ufo_module.Font()
    ufo1.layers["public.default"].lib["layerLibKey1"] = "ufo1 layerLibValue1"
    layer = ufo1.newLayer("sketches")
    layer.lib["layerLibKey2"] = "ufo1 layerLibValue2"
    # layers won't roundtrip if they contain no glyph, except for the default
    layer.newGlyph("bob")
    ufo2 = ufo_module.Font()
    ufo2.layers["public.default"].lib["layerLibKey1"] = "ufo2 layerLibValue1"
    layer = ufo2.newLayer("sketches")
    layer.lib["layerLibKey2"] = "ufo2 layerLibValue2"
    layer.newGlyph("bob")

    font = to_glyphs([ufo1, ufo2])

    default_layer_key = GLYPHLIB_PREFIX + "layerLib.public.default"
    sketches_layer_key = GLYPHLIB_PREFIX + "layerLib.sketches"
    assert default_layer_key not in font.userData
    assert sketches_layer_key not in font.userData
    assert font.masters[0].userData[default_layer_key] == {
        "layerLibKey1": "ufo1 layerLibValue1"
    }
    assert font.masters[0].userData[sketches_layer_key] == {
        "layerLibKey2": "ufo1 layerLibValue2"
    }
    assert font.masters[1].userData[default_layer_key] == {
        "layerLibKey1": "ufo2 layerLibValue1"
    }
    assert font.masters[1].userData[sketches_layer_key] == {
        "layerLibKey2": "ufo2 layerLibValue2"
    }

    (ufo1, ufo2) = to_ufos(font, minimal=False)

    assert ufo1.layers["public.default"].lib["layerLibKey1"] == "ufo1 layerLibValue1"
    assert "layerLibKey1" not in ufo1.layers["sketches"].lib
    assert ufo1.layers["sketches"].lib["layerLibKey2"] == "ufo1 layerLibValue2"
    assert "layerLibKey2" not in ufo1.layers["public.default"].lib
    assert ufo2.layers["public.default"].lib["layerLibKey1"] == "ufo2 layerLibValue1"
    assert "layerLibKey1" not in ufo2.layers["sketches"].lib
    assert ufo2.layers["sketches"].lib["layerLibKey2"] == "ufo2 layerLibValue2"
    assert "layerLibKey2" not in ufo2.layers["public.default"].lib


def test_layer_lib_in_font_user_data(ufo_module):
    font = classes.GSFont()
    font.masters.append(classes.GSFontMaster())
    font.masters.append(classes.GSFontMaster())
    glyph = classes.GSGlyph(name="a")
    font.glyphs.append(glyph)
    layer = classes.GSLayer()
    layer.layerId = font.masters[0].id
    layer.width = 0
    glyph.layers.append(layer)
    layer = classes.GSLayer()
    layer.layerId = font.masters[1].id
    layer.width = 0
    glyph.layers.append(layer)

    font.userData[GLYPHLIB_PREFIX + "layerLib.public.default"] = {
        "layerLibKey": "layerLibValue"
    }

    ufo1, ufo2 = to_ufos(font)
    assert "layerLibKey" in ufo1.layers["public.default"].lib
    assert ufo1.layers["public.default"].lib["layerLibKey"] == "layerLibValue"
    assert "layerLibKey" in ufo2.layers["public.default"].lib
    assert ufo2.layers["public.default"].lib["layerLibKey"] == "layerLibValue"


def test_glyph_user_data_into_ufo_lib():
    font = classes.GSFont()
    font.masters.append(classes.GSFontMaster())
    glyph = classes.GSGlyph("a")
    glyph.userData["glyphUserDataKey"] = "glyphUserDataValue"
    font.glyphs.append(glyph)
    layer = classes.GSLayer()
    layer.layerId = font.masters[0].id
    glyph.layers.append(layer)

    (ufo,) = to_ufos(font)

    assert ufo.lib[GLYPHLIB_PREFIX + "glyphUserData.a"] == {
        "glyphUserDataKey": "glyphUserDataValue"
    }

    font = to_glyphs([ufo])

    assert font.glyphs["a"].userData["glyphUserDataKey"] == "glyphUserDataValue"


@pytest.mark.parametrize("minimal", [True, False])
def test_math_user_data_into_ufo_lib(datadir, minimal, caplog):
    font = classes.GSFont(str(datadir.join("Math.glyphs")))

    ufos = to_ufos(font, minimal=minimal)

    math_glyphs = ["parenleft", "parenright"]

    for ufo, master in zip(ufos, font.masters):
        assert ufo.lib[GLYPHS_MATH_EXTENDED_SHAPE_KEY] == math_glyphs
        assert (
            ufo.lib[GLYPHS_MATH_CONSTANTS_KEY]
            == master.userData[GLYPHS_MATH_CONSTANTS_KEY]
        )
        for name in math_glyphs:
            for key in {"vVariants", "hVariants"}:
                result = ufo[name].lib[GLYPHS_MATH_VARIANTS_KEY].get(key)
                expected = font.glyphs[name].userData[GLYPHS_MATH_VARIANTS_KEY].get(key)
                assert result == expected
            for key in {"vAssembly", "hAssembly"}:
                result = ufo[name].lib[GLYPHS_MATH_VARIANTS_KEY].get(key)
                expected = (
                    font.glyphs[name]
                    .layers[master.id]
                    .userData[GLYPHS_MATH_VARIANTS_KEY]
                    .get(key)
                )
                assert result == expected

    font2 = to_glyphs(ufos)
    assert "already has different 'vVariants'" not in caplog.text

    for master, ufo in zip(font2.masters, ufos):
        assert font2.userData[GLYPHS_MATH_EXTENDED_SHAPE_KEY] is None
        assert font2.userData[GLYPHS_MATH_CONSTANTS_KEY] is None
        assert (
            master.userData[GLYPHS_MATH_CONSTANTS_KEY]
            == ufo.lib[GLYPHS_MATH_CONSTANTS_KEY]
        )
        for name in math_glyphs:
            for key in {"vVariants", "hVariants"}:
                result = font2.glyphs[name].userData[GLYPHS_MATH_VARIANTS_KEY].get(key)
                expected = ufo[name].lib[GLYPHS_MATH_VARIANTS_KEY].get(key)
                assert result == expected
            for key in {"vAssembly", "hAssembly"}:
                result = (
                    font2.glyphs[name]
                    .layers[master.id]
                    .userData[GLYPHS_MATH_VARIANTS_KEY]
                    .get(key)
                )
                expected = ufo[name].lib[GLYPHS_MATH_VARIANTS_KEY].get(key)
                assert result == expected


@pytest.mark.parametrize("minimal", [True, False])
def test_math_user_data_into_ufo_lib_warn(datadir, minimal, caplog):
    import copy

    font = classes.GSFont(str(datadir.join("Math.glyphs")))

    ufos = to_ufos(font, minimal=minimal)
    ufos[0]["parenleft"] = copy.deepcopy(ufos[0]["parenleft"])
    ufos[0]["parenleft"].lib[GLYPHS_MATH_VARIANTS_KEY]["vVariants"].pop()

    assert (
        ufos[0]["parenleft"].lib[GLYPHS_MATH_VARIANTS_KEY]["vVariants"]
        != ufos[1]["parenleft"].lib[GLYPHS_MATH_VARIANTS_KEY]["vVariants"]
    )

    to_glyphs(ufos)

    assert any(record.levelname == "WARNING" for record in caplog.records)
    assert (
        "Glyph 'parenleft' already has different 'vVariants' in "
        "userData['com.nagwa.MATHPlugin.variants']. Overwriting it." in caplog.text
    )


def test_glif_lib_equivalent_to_layer_user_data(ufo_module):
    ufo = ufo_module.Font()
    # This glyph is in the `public.default` layer
    a = ufo.newGlyph("a")
    a.lib["glifLibKeyA"] = "glifLibValueA"
    customLayer = ufo.newLayer("middleground")
    # "a" is in both layers
    customLayer.newGlyph("a")
    # "b" is only in the second layer
    b = customLayer.newGlyph("b")
    b.lib["glifLibKeyB"] = "glifLibValueB"

    font = to_glyphs([ufo])

    for layer_id in font.glyphs["a"].layers.keys():
        layer = font.glyphs["a"].layers[layer_id]
        if layer.layerId == font.masters[0].id:
            default_layer = layer
        else:
            middleground = layer
    assert default_layer.userData["glifLibKeyA"] == "glifLibValueA"
    assert "glifLibKeyA" not in middleground.userData.keys()

    for layer_id in font.glyphs["b"].layers.keys():
        layer = font.glyphs["b"].layers[layer_id]
        if layer.layerId == font.masters[0].id:
            default_layer = layer
        else:
            middleground = layer
    assert "glifLibKeyB" not in default_layer.userData.keys()
    assert middleground.userData["glifLibKeyB"] == "glifLibValueB"

    (ufo,) = to_ufos(font, minimal=False)

    assert ufo["a"].lib["glifLibKeyA"] == "glifLibValueA"
    assert "glifLibKeyA" not in ufo.layers["middleground"]["a"]
    assert ufo.layers["middleground"]["b"].lib["glifLibKeyB"] == "glifLibValueB"


def test_node_user_data_into_glif_lib():
    font = classes.GSFont()
    master = classes.GSFontMaster()
    master.id = "M1"
    font.masters.append(master)
    glyph = classes.GSGlyph("a")
    layer = classes.GSLayer()
    layer.layerId = "M1"
    layer.associatedMasterId = "M1"
    glyph.layers.append(layer)
    font.glyphs.append(glyph)
    path = classes.GSPath()
    layer.paths.append(path)
    node1 = classes.GSNode()
    node1.userData["nodeUserDataKey1"] = "nodeUserDataValue1"
    node1.userData["UFO.identifier"] = "id1"
    node1.name = "node1"
    node2 = classes.GSNode()
    node2.userData["nodeUserDataKey2"] = "nodeUserDataValue2"
    node2.userData["UFO.identifier"] = "id2"
    node2.name = "node2"
    path.nodes.append(classes.GSNode())
    path.nodes.append(node1)
    path.nodes.append(classes.GSNode())
    path.nodes.append(classes.GSNode())
    path.nodes.append(node2)

    (ufo,) = to_ufos(font, minimize_glyphs_diffs=True)

    assert ufo["a"].lib[GLYPHLIB_PREFIX + "nodeUserData.0.1"] == {
        "nodeUserDataKey1": "nodeUserDataValue1"
    }
    assert ufo["a"][0][2].name == "node1"
    assert ufo["a"][0][2].identifier == "id1"
    assert ufo["a"].lib[GLYPHLIB_PREFIX + "nodeUserData.0.4"] == {
        "nodeUserDataKey2": "nodeUserDataValue2"
    }
    assert ufo["a"][0][0].name == "node2"
    assert ufo["a"][0][0].identifier == "id2"

    font = to_glyphs([ufo])

    path = font.glyphs["a"].layers["M1"].paths[0]
    assert path.nodes[1].userData["nodeUserDataKey1"] == "nodeUserDataValue1"
    assert path.nodes[1].name == "node1"
    assert path.nodes[1].userData["UFO.identifier"] == "id1"
    assert path.nodes[4].userData["nodeUserDataKey2"] == "nodeUserDataValue2"
    assert path.nodes[4].name == "node2"
    assert path.nodes[4].userData["UFO.identifier"] == "id2"


def test_lib_data_types(tmpdir, ufo_module):
    # Test the roundtrip of a few basic types both at the top level and in a
    # nested object.
    data = OrderedDict(
        {
            "boolean": True,
            "smooth": False,
            "integer": 1,
            "float": 0.5,
            "array": [],
            "dict": {},
        }
    )
    ufo = ufo_module.Font()
    a = ufo.newGlyph("a")
    for key, value in data.items():
        a.lib[key] = value
        a.lib["nestedDict"] = dict(data)
        a.lib["nestedArray"] = list(data.values())
        a.lib["crazyNesting"] = [{"a": [{"b": [dict(data)]}]}]

    font = to_glyphs([ufo])

    # FIXME: This test will stop working if the font is written and read back,
    # because the file format of Glyphs does not make a difference between
    # `True` (bool) and `1` (int).
    # filename = os.path.join(str(tmpdir), 'font.glyphs')
    # font.save(filename)
    # font = classes.GSFont(filename)

    (ufo,) = to_ufos(font)

    for index, (key, value) in enumerate(data.items()):
        assert value == ufo["a"].lib[key]
        assert value == ufo["a"].lib["nestedDict"][key]
        assert value == ufo["a"].lib["nestedArray"][index]
        assert value == ufo["a"].lib["crazyNesting"][0]["a"][0]["b"][0][key]
        assert type(value) is type(ufo["a"].lib[key])  # noqa: E721
        assert type(value) is type(ufo["a"].lib["nestedDict"][key])  # noqa: E721
        assert type(value) is type(ufo["a"].lib["nestedArray"][index])  # noqa: E721
        assert type(value) is type(  # noqa: E721
            ufo["a"].lib["crazyNesting"][0]["a"][0]["b"][0][key]
        )



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re

from glyphsLib.types import Point
from glyphsLib.builder.constants import LOCKED_GUIDE_NAME_SUFFIX

COLOR_NAME_SUFFIX = " [%s]"
COLOR_NAME_RE = re.compile(r"^.*( \[([0-1.,eE+-]+)\])$")
IDENTIFIER_NAME_SUFFIX = " [#%s]"
IDENTIFIER_NAME_RE = re.compile(r"^.*( \[#([^\]]+)\])$")


def to_ufo_guidelines(self, ufo_obj, glyphs_obj):
    """Set guidelines."""
    guidelines = glyphs_obj.guides
    if not guidelines:
        return
    new_guidelines = []
    for guideline in guidelines:
        new_guideline = {}
        x, y = guideline.position
        angle = guideline.angle % 360
        if _is_vertical(x, y, angle):
            new_guideline["x"] = x
        elif _is_horizontal(x, y, angle):
            new_guideline["y"] = y
        else:
            new_guideline["x"] = x
            new_guideline["y"] = y
            new_guideline["angle"] = angle
        name = guideline.name
        if name is not None:
            # Identifier
            m = IDENTIFIER_NAME_RE.match(name)
            if m:
                new_guideline["identifier"] = m.group(2)
                name = name[: -len(m.group(1))]
            # Color
            m = COLOR_NAME_RE.match(name)
            if m:
                new_guideline["color"] = m.group(2)
                name = name[: -len(m.group(1))]
        if guideline.locked:
            name = (name or "") + LOCKED_GUIDE_NAME_SUFFIX
        if name:
            new_guideline["name"] = name
        new_guidelines.append(new_guideline)
    ufo_obj.guidelines = new_guidelines


def to_glyphs_guidelines(self, ufo_obj, glyphs_obj):
    """Set guidelines."""
    if not ufo_obj.guidelines:
        return
    for guideline in ufo_obj.guidelines:
        new_guideline = self.glyphs_module.GSGuide()
        name = guideline.name
        # Locked
        if name is not None and name.endswith(LOCKED_GUIDE_NAME_SUFFIX):
            name = name[: -len(LOCKED_GUIDE_NAME_SUFFIX)]
            new_guideline.locked = True
        if guideline.color:
            name = (name or "") + COLOR_NAME_SUFFIX % str(guideline.color)
        if guideline.identifier:
            name = (name or "") + IDENTIFIER_NAME_SUFFIX % guideline.identifier
        new_guideline.name = name
        new_guideline.position = Point(guideline.x or 0, guideline.y or 0)
        if guideline.angle is not None:
            new_guideline.angle = guideline.angle % 360
        elif _is_vertical(guideline.x, guideline.y, None):
            new_guideline.angle = 90
        glyphs_obj.guides.append(new_guideline)


def _is_vertical(x, y, angle):
    return (y is None or y == 0) and (angle is None or angle == 90 or angle == 270)


def _is_horizontal(x, y, angle):
    return (x is None or x == 0) and (angle is None or angle == 0 or angle == 180)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .constants import ANNOTATIONS_LIB_KEY
from glyphsLib.types import Point


def to_ufo_annotations(self, ufo_glyph, layer):
    try:
        value = layer.annotations
    except KeyError:
        return
    annotations = []
    for an in list(value.values()):
        annot = {}
        for attr in ["angle", "position", "text", "type", "width"]:
            val = getattr(an, attr, None)
            if attr == "position" and val:
                val = list(val)
            if val:
                annot[attr] = val
        annotations.append(annot)

    if annotations:
        ufo_glyph.lib[ANNOTATIONS_LIB_KEY] = annotations


def to_glyphs_annotations(self, ufo_glyph, layer):
    if ANNOTATIONS_LIB_KEY not in ufo_glyph.lib:
        return

    for annot in ufo_glyph.lib[ANNOTATIONS_LIB_KEY]:
        annotation = self.glyphs_module.GSAnnotation()
        for attr in ["angle", "position", "text", "type", "width"]:
            if attr in annot and annot[attr]:
                if attr == "position":
                    # annot['position'] can be either "{1, 2}" or (1, 2)
                    position = Point(annot["position"])
                    annotation.position = position
                else:
                    setattr(annotation, attr, annot[attr])
        layer.annotations.append(annotation)
